// ==============================================================
//                 ORBITER MODULE: SDO
//                  
//
// sdo.cpp
// Control module for sdo vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================

const double PA_ISP = 2800;
const double PB_MAXRCSTH = 20;
const double PB_MAXRCSTH20 = 500;

const double ANTENNA_DEPLOY_SPEED = 0.1;
const double ROTA_OPERATING_SPEED = 0.01;
const double ROTB_OPERATING_SPEED = 0.04;
const double ROTC_OPERATING_SPEED = 0.04;
const double SOLP_DEPLOY_SPEED = 0.5;
const double LID_DEPLOY_SPEED = 1.0;

class sdo : public VESSEL3 {
public:
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } ant_status, solp_status, lid_status;
	sdo(OBJHANDLE hVessel, int flightmodel)
		: VESSEL3(hVessel, flightmodel) {}
	void clbkSetClassCaps(FILEHANDLE cfg);
	void clbkPostStep(double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState(FILEHANDLE scn);
	void clbkLoadStateEx(FILEHANDLE scn, void *status);
	void DefineAnimations(void);
	void RevertAntenna(void);
	void RevertSolp(void);
	void RevertLid(void);
	void AntEarth(void);
	void DoCalculations(void);
	bool clbkDrawHUD(int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);

	THRUSTER_HANDLE th_rcs[17], th_group[4];
	PROPELLANT_HANDLE rcspr;

	ANIMATIONCOMPONENT_HANDLE Section1, Section2, Section3, Section4;
private:
	UINT anim_ant, ant_rota, ant_rotb, ant_rotc, anim_solp, anim_lid;
	double ant_proc, rota_proc, rotb_proc, rotc_proc, solp_proc, lid_proc;
	double stepsize;
	double delta_long;
	double deg_long_to_target_e;
	double deg_long_to_target_w;
	double orbits_required;
	double mean_earth_long;

	int a;
	int b;
	int c;

};

void sdo::DefineAnimations(void)
{
	// 1. ant
	static UINT HiGainAntA1Grp[3] = { 7,14,18 };
	static MGROUP_ROTATE HiGainAnt1(0, HiGainAntA1Grp, 3, _V(0, 1.161, -2.204), _V(-1, 0, 0), float(PI / 2));

	static UINT HiGainAntB1Grp[3] = { 6,12,21 };
	static MGROUP_ROTATE HiGainAnt2(0, HiGainAntB1Grp, 3, _V(0, -1.165, -2.204), _V(1, 0, 0), float(PI / 2));

	anim_ant = CreateAnimation(0);
	Section1 = AddAnimationComponent(anim_ant, 0, 1, &HiGainAnt1);
	Section2 = AddAnimationComponent(anim_ant, 0, 1, &HiGainAnt2);

	// 2. rota
	static UINT HiGainAntA2Grp[] = { 0 };
	static MGROUP_ROTATE HiGainAnt3(0, HiGainAntA2Grp, 0, _V(0, 1.161, 0), _V(0, 0, -1), float(2 * PI));

	static UINT HiGainAntB2Grp[] = { 0 };
	static MGROUP_ROTATE HiGainAnt4(0, HiGainAntB2Grp, 0, _V(0, -1.165, 0), _V(0, 0, 1), float(2 * PI));

	ant_rota = CreateAnimation(0);
	Section3 = AddAnimationComponent(ant_rota, 0, 1, &HiGainAnt3, Section1);
	Section4 = AddAnimationComponent(ant_rota, 0, 1, &HiGainAnt4, Section2);

	// 3. rotb
	static UINT HiGainAntA3Grp[4] = { 13,15,16,17 };
	static MGROUP_ROTATE HiGainAnt5(0, HiGainAntA3Grp, 4, _V(0, 1.161, -0.065), _V(1, 0, 0), float(0.5*PI));

	ant_rotb = CreateAnimation(0);
	AddAnimationComponent(ant_rotb, 0, 1, &HiGainAnt5, Section3);

	// 4. rotc
	static UINT HiGainAntB3Grp[4] = { 8,9,10,11 };
	static MGROUP_ROTATE HiGainAnt6(0, HiGainAntB3Grp, 4, _V(0, -1.165, -0.065), _V(-1, 0, 0), float(0.5*PI));

	ant_rotc = CreateAnimation(0);
	AddAnimationComponent(ant_rotc, 0, 1, &HiGainAnt6, Section4);

	// 5. solp
	static UINT HiGainAnt4Grp[1] = { 19 };
	static MGROUP_ROTATE solp1(0, HiGainAnt4Grp, 1, _V(-1.095, 0, 0.795), _V(0, 1, 0), float(0.5*PI));

	static UINT HiGainAnt5Grp[1] = { 20 };
	static MGROUP_ROTATE solp2(0, HiGainAnt5Grp, 1, _V(1.095, 0, 0.795), _V(0, -1, 0), float(0.5*PI));

	anim_solp = CreateAnimation(0);
	AddAnimationComponent(anim_solp, 0, 1, &solp1);
	AddAnimationComponent(anim_solp, 0, 1, &solp2);

	// 6. lids
	static UINT HiGainAnt7Grp[1] = { 4 };
	static MGROUP_ROTATE HiGainAnt7(0, HiGainAnt7Grp, 1, _V(0, 1.067, 3.049), _V(1, 0, 0), float(PI));

	anim_lid = CreateAnimation(0);
	AddAnimationComponent(anim_lid, 0, 1, &HiGainAnt7);
}

void sdo::RevertAntenna(void)
{
	ant_status = ((ant_status == DOOR_CLOSED || ant_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void sdo::RevertSolp(void)
{
	solp_status = ((solp_status == DOOR_CLOSED || solp_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}

void sdo::RevertLid(void)
{
	lid_status = ((lid_status == DOOR_CLOSED || lid_status == DOOR_CLOSING) ?
		DOOR_OPENING : DOOR_CLOSING);
}



// ==============================================================
// sdo class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void sdo::clbkSetClassCaps(FILEHANDLE cfg)
{

	ant_proc = 0.0;
	solp_proc = 0.0;
	rota_proc = 0.0;
	rotb_proc = 0.0;
	rotc_proc = 0.0;
	lid_proc = 0.0;
	stepsize = 0.0;
	a = 0;
	b = 0;
	c = 0;
	ant_status = DOOR_CLOSED;
	solp_status = DOOR_CLOSED;
	lid_status = DOOR_CLOSED;
	DefineAnimations();
	// physical specs
	SetCW(1, 1, 1, 1);
	SetRotDrag(_V(1, 1, 1));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale(1e-3);
		SetBankMomentScale(1e-3);
	}
	SetTrimScale(0.1);
	SetPMI(_V(2, 2, 2));
	SetCrossSections(_V(10, 10, 6.5));
	SetSize(4);
	SetEmptyMass(1650);
	SetMeshVisibilityMode((AddMesh("sdo/sdo")), MESHVIS_ALWAYS);
	SetCameraOffset(_V(0, 2, -5));
	SetTouchdownPoints(_V(0, 1, 0), _V(-1, -1, 0), _V(1, -1, 0));

	// propellant resources
	const double PA_FUELMASS = 1450;
	double rcsmass = 1450;

	rcspr = CreatePropellantResource(PA_FUELMASS, rcsmass);
	// ***************** thruster definitions *******************

	// pitchup
	th_rcs[0] = CreateThruster(_V(0, -1, 0), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[1] = CreateThruster(_V(0, 1, 0), _V(0, 0, -1), PB_MAXRCSTH, rcspr, PA_ISP);
	// pitchdown
	th_rcs[2] = CreateThruster(_V(0, -1, 0), _V(0, 0, -1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[3] = CreateThruster(_V(0, 1, 0), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	//bankleft
	th_rcs[4] = CreateThruster(_V(0, -1, 0), _V(1, 0, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[5] = CreateThruster(_V(0, 1, 0), _V(-1, 0, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	//bankright
	th_rcs[6] = CreateThruster(_V(0, -1, 0), _V(-1, 0, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[7] = CreateThruster(_V(0, 1, 0), _V(1, 0, 0), PB_MAXRCSTH, rcspr, PA_ISP);
	//yawleft
	th_rcs[8] = CreateThruster(_V(1, 0, 0), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[9] = CreateThruster(_V(-1, 0, 0), _V(0, 0, -1), PB_MAXRCSTH, rcspr, PA_ISP);
	//yawright
	th_rcs[10] = CreateThruster(_V(1, 0, 0), _V(0, 0, -1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[11] = CreateThruster(_V(-1, 0, 0), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	//attforward
	th_rcs[12] = CreateThruster(_V(1, -1, -1), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[13] = CreateThruster(_V(1, 1, -1), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[14] = CreateThruster(_V(-1, -1, -1), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	th_rcs[15] = CreateThruster(_V(-1, 1, -1), _V(0, 0, 1), PB_MAXRCSTH, rcspr, PA_ISP);
	//main
	th_rcs[16] = CreateThruster(_V(0, 0, -1), _V(0, 0, 1), PB_MAXRCSTH20, rcspr, PA_ISP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[1];
	CreateThrusterGroup(th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[2];
	th_group[1] = th_rcs[3];
	CreateThrusterGroup(th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[4];
	th_group[1] = th_rcs[5];
	CreateThrusterGroup(th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[6];
	th_group[1] = th_rcs[7];
	CreateThrusterGroup(th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[9];
	CreateThrusterGroup(th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[10];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup(th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[12];
	th_group[1] = th_rcs[13];
	th_group[2] = th_rcs[14];
	th_group[3] = th_rcs[15];
	CreateThrusterGroup(th_group, 4, THGROUP_ATT_FORWARD);

	CreateThrusterGroup(th_rcs + 16, 1, THGROUP_MAIN);

	SURFHANDLE exhaust = oapiRegisterExhaustTexture("exhaust_atrcs");

	AddExhaust(th_rcs[0], 0.5, 0.05, _V(0.977, -0.894, -2.475), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[0], 0.5, 0.05, _V(0.977, -0.811, -2.453), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[0], 0.5, 0.05, _V(-0.977, -0.894, -2.475), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[0], 0.5, 0.05, _V(-0.977, -0.811, -2.453), _V(0.259, 0.259, -0.634), (exhaust));

	AddExhaust(th_rcs[2], 0.5, 0.05, _V(0.977, 0.894, -2.475), _V(-0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[2], 0.5, 0.05, _V(0.977, 0.811, -2.453), _V(-0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[2], 0.5, 0.05, _V(-0.977, 0.894, -2.475), _V(0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[2], 0.5, 0.05, _V(-0.977, 0.811, -2.453), _V(0.259, -0.259, -0.634), (exhaust));

	AddExhaust(th_rcs[6], 0.5, 0.05, _V(0.977, -0.894, -2.475), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[6], 0.5, 0.05, _V(0.977, -0.811, -2.453), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[6], 0.5, 0.05, _V(-0.977, 0.894, -2.475), _V(0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[6], 0.5, 0.05, _V(-0.977, 0.811, -2.453), _V(0.259, -0.259, -0.634), (exhaust));

	AddExhaust(th_rcs[4], 0.5, 0.05, _V(-0.977, -0.894, -2.475), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[4], 0.5, 0.05, _V(-0.977, -0.811, -2.453), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[4], 0.5, 0.05, _V(0.977, 0.894, -2.475), _V(-0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[4], 0.5, 0.05, _V(0.977, 0.811, -2.453), _V(-0.259, -0.259, -0.634), (exhaust));

	AddExhaust(th_rcs[8], 0.5, 0.05, _V(0.977, -0.894, -2.475), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[8], 0.5, 0.05, _V(0.977, -0.811, -2.453), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[8], 0.5, 0.05, _V(0.977, 0.894, -2.475), _V(-0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[8], 0.5, 0.05, _V(0.977, 0.811, -2.453), _V(-0.259, -0.259, -0.634), (exhaust));

	AddExhaust(th_rcs[10], 0.5, 0.05, _V(-0.977, -0.894, -2.475), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[10], 0.5, 0.05, _V(-0.977, -0.811, -2.453), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[10], 0.5, 0.05, _V(-0.977, 0.894, -2.475), _V(0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[10], 0.5, 0.05, _V(-0.977, 0.811, -2.453), _V(0.259, -0.259, -0.634), (exhaust));

	AddExhaust(th_rcs[12], 0.5, 0.05, _V(0.977, -0.894, -2.475), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(0.977, -0.811, -2.453), _V(-0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(-0.977, -0.894, -2.475), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(-0.977, -0.811, -2.453), _V(0.259, 0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(0.977, 0.894, -2.475), _V(-0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(0.977, 0.811, -2.453), _V(-0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(-0.977, 0.894, -2.475), _V(0.259, -0.259, -0.634), (exhaust));
	AddExhaust(th_rcs[12], 0.5, 0.05, _V(-0.977, 0.811, -2.453), _V(0.259, -0.259, -0.634), (exhaust));

	SURFHANDLE exhaust2 = oapiRegisterExhaustTexture("exhaust");
	AddExhaust(th_rcs[16], 1.5, 0.15, _V(0, 0, -2.63), _V(0, 0, -1), (exhaust2));
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit(OBJHANDLE hvessel, int flightmodel)
{
	return new sdo(hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit(VESSEL *vessel)
{
	if (vessel) delete (sdo*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void sdo::clbkPostCreation()
{
	double rcsprmass = 1450;

	if (GetPropellantMass(rcspr) < 0.01 && oapiGetSimTime() > 0.1)SetPropellantMass(rcspr, rcsprmass);

}
void sdo::clbkSaveState(FILEHANDLE scn)
{
	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState(scn);

	// custom parameters
	sprintf(cbuf, "%d %0.4f %0.4f %0.4f %0.4f", ant_status, ant_proc, rota_proc, rotb_proc, rotc_proc);
	oapiWriteScenario_string(scn, "ANT", cbuf);
	sprintf(cbuf, "%d %0.4f", solp_status, solp_proc);
	oapiWriteScenario_string(scn, "SOLP", cbuf);
	sprintf(cbuf, "%d %0.4f", lid_status, lid_proc);
	oapiWriteScenario_string(scn, "LID", cbuf);
}
void sdo::clbkLoadStateEx(FILEHANDLE scn, void *vs)
{
	char *line;
	while (oapiReadScenario_nextline(scn, line)) {
		if (!strnicmp(line, "ANT", 3)) {
			sscanf(line + 3, "%d%lf%lf%lf%lf", &ant_status, &ant_proc, &rota_proc, &rotb_proc, &rotc_proc);
		}
		else if (!strnicmp(line, "SOLP", 4)) {
			sscanf(line + 4, "%d%lf", &solp_status, &solp_proc);
		}
		else if (!strnicmp(line, "LID", 3)) {
			sscanf(line + 4, "%d%lf", &lid_status, &lid_proc);
		}
		else {
			ParseScenarioLineEx(line, vs);
			// unrecognised option - pass to Orbiter's generic parser
		}
	}
	SetAnimation(anim_ant, ant_proc);
	SetAnimation(ant_rota, rota_proc);
	SetAnimation(ant_rotb, rotb_proc);
	SetAnimation(ant_rotc, rotc_proc);
	SetAnimation(anim_solp, solp_proc);
	SetAnimation(anim_lid, lid_proc);
}
int sdo::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	if (KEYMOD_SHIFT(kstate)) {

	}
	else { // unmodified keys
		switch (key) {

		case OAPI_KEY_G: // deploy ant and solp
			if (ant_status == 0)
			{
				RevertAntenna();
			}
			if (solp_status == 0)
			{
				RevertSolp();
			}

			return 1;

		case OAPI_KEY_K: // open lids
			if (lid_status == 0)
			{
				RevertLid();
			}
			return 1;
		}
	}
	return 0;
}

void sdo::DoCalculations(void)
{
	char cbuf[7] = "Earth";
	OBJHANDLE hRef = oapiGetObjectByName(cbuf);

	ELEMENTS el;
	ORBITPARAM prm;
	double date = oapiGetSimMJD();
	GetElements(hRef, el, &prm, date, FRAME_EQU);

	double earth_rot_angle = oapiGetPlanetCurrentRotation(hRef);

	if (earth_rot_angle > (2 * PI))earth_rot_angle = earth_rot_angle - (2 * PI);

	mean_earth_long = DEG * (prm.MnL - earth_rot_angle);

	if (mean_earth_long < 0) mean_earth_long = mean_earth_long + 360;

	double period = prm.T;

	delta_long = 360 * ((86164.09 - period) / 86164.09);

	deg_long_to_target_e = 258 - mean_earth_long;

	if (deg_long_to_target_e < 0) deg_long_to_target_e = 360 + deg_long_to_target_e;

	deg_long_to_target_w = 360 - deg_long_to_target_e;

	if (delta_long > 0) orbits_required = deg_long_to_target_e / delta_long;
	else if (delta_long < 0) orbits_required = -deg_long_to_target_w / delta_long;
}


void sdo::AntEarth(void) {

	//ant earth rota

	{
		double Ant_angle;
		double Earth_angle;
		char cbuf[7] = "Earth";
		OBJHANDLE hRef = oapiGetObjectByName(cbuf);
		VECTOR3 rearthloc;
		VECTOR3 earthglobal;
		oapiGetGlobalPos(hRef, &earthglobal);
		Global2Local(earthglobal, rearthloc);
		Ant_angle = (PI*((rota_proc * 2) + 0.0));
		Earth_angle = atan2(rearthloc.x, -rearthloc.z);
		if (Earth_angle < 0) Earth_angle = Earth_angle + (2 * PI);
		if (Earth_angle > (2 * PI)) Earth_angle = Earth_angle - (2 * PI);

		if ((Ant_angle - Earth_angle) > (-ROTA_OPERATING_SPEED * stepsize * 2 * PI) && (Ant_angle - Earth_angle) < (ROTA_OPERATING_SPEED*stepsize * 2 * PI))
		{
			rota_proc = Earth_angle / (2 * PI);
			a = 0;
		}

		else if ((Ant_angle - Earth_angle) > 0)
		{
			if (Ant_angle - Earth_angle < PI) a = 2;
			if (Ant_angle - Earth_angle > PI) a = 1;
		}

		else if ((Ant_angle - Earth_angle) < 0)
		{
			if (Ant_angle - Earth_angle > -PI) a = 1;
			if (Ant_angle - Earth_angle < -PI) a = 2;
		}

		double db = stepsize * ROTA_OPERATING_SPEED;

		if (a == 2) rota_proc = (rota_proc - db);
		else if (a == 1) rota_proc = (rota_proc + db);

		if (rota_proc > 1) rota_proc = (rota_proc - 1);
		else if (rota_proc < 0) rota_proc = (rota_proc + 1);
		SetAnimation(ant_rota, rota_proc);

		////////////////////////////////ant earth rotb
		double Ant_angleb;
		double Ant_anglec;
		double yposaxis_angle;
		double ynegaxis_angle;

		yposaxis_angle = acos(dotp(_V(0, 1, 0), rearthloc) / length(rearthloc));
		if (yposaxis_angle > (0.5*PI)) yposaxis_angle = (0.5*PI);

		Ant_angleb = 0.5*PI*(1 - rotb_proc);

		if ((Ant_angleb - yposaxis_angle) > (-ROTB_OPERATING_SPEED * stepsize*0.5*PI) && (Ant_angleb - yposaxis_angle) < (ROTB_OPERATING_SPEED*stepsize*0.5*PI))
		{
			rotb_proc = 1 - (yposaxis_angle / (0.5*PI));
			b = 0;
		}

		else if ((Ant_angleb - yposaxis_angle) > 0) b = 1;
		else if ((Ant_angleb - yposaxis_angle) < 0) b = 2;

		double da = stepsize * ROTB_OPERATING_SPEED;

		if (b == 2) rotb_proc = (rotb_proc - da);
		else if (b == 1) rotb_proc = (rotb_proc + da);

		if (rotb_proc > 1) rotb_proc = 1;
		else if (rotb_proc < 0) rotb_proc = 0;

		SetAnimation(ant_rotb, rotb_proc);

		////////// ant rot c

		ynegaxis_angle = acos(dotp(_V(0, -1, 0), rearthloc) / length(rearthloc));
		if (ynegaxis_angle > (0.5*PI))ynegaxis_angle = (0.5*PI);

		Ant_anglec = 0.5*PI*(1 - rotc_proc);

		if ((Ant_anglec - ynegaxis_angle) > (-ROTC_OPERATING_SPEED * stepsize*0.5*PI) && (Ant_anglec - ynegaxis_angle) < (ROTC_OPERATING_SPEED*stepsize*0.5*PI))
		{
			rotc_proc = 1 - (ynegaxis_angle / (0.5*PI));
			c = 0;
		}

		else if ((Ant_anglec - ynegaxis_angle) > 0) c = 1;
		else if ((Ant_anglec - ynegaxis_angle) < 0) c = 2;

		double dc = stepsize * ROTC_OPERATING_SPEED;

		if (c == 2) rotc_proc = (rotc_proc - dc);
		else if (c == 1) rotc_proc = (rotc_proc + dc);

		if (rotc_proc > 1) rotc_proc = 1;
		else if (rotc_proc < 0) rotc_proc = 0;

		SetAnimation(ant_rotc, rotc_proc);
	}
}

bool sdo::clbkDrawHUD(int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw = (int)(16 * sw / 1024);
	int hlw = (int)(lw / 2);

	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds / dsw;

	if (sc_ratio < 0.7284)
	{
		roxl = (lw * 10);
		royl = (int)(-88 * d);
	}

	int wd = (int)(152 * d);
	int wc = (int)(168 * d);
	int w0 = (int)(184 * d);
	int w1 = (int)(200 * d);
	int w2 = (int)(216 * d);
	int w3 = (int)(232 * d);
	int w4 = (int)(248 * d);
	int w5 = (int)(264 * d);
	int w6 = (int)(280 * d);
	int w7 = (int)(296 * d);

	// draw the default HUD
	VESSEL3::clbkDrawHUD(mode, hps, skp);

	{
		double m1 = GetEmptyMass();
		double m2 = GetPropellantMass(rcspr);
		double test = ((m1 + m2) / m1);
		double dv = (2800 * (log(test)));
		skp->SetTextColor(0x0066FF66);
		char kbuf[256];
		sprintf(kbuf, "Remaining dV = %0.2f m/s", dv);
		skp->Text((10 + roxl), (wc + royl), kbuf, strlen(kbuf));
	}

	if (ant_status == 0)
	{
		skp->SetTextColor(0x0066FFFF);
		char abuf[256];
		sprintf(abuf, "G = Deploy HGA + Solar Panels");
		skp->Text((10 + roxl), (w0 + royl), abuf, strlen(abuf));
	}
	else
	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "HGA and Solar Panels Deployed");
		skp->Text((10 + roxl), (w0 + royl), bbuf, strlen(bbuf));
	}

	if (lid_status == 0)
	{
		skp->SetTextColor(0x0066FFFF);
		char abuf[256];
		sprintf(abuf, "K = Open AIA Covers");
		skp->Text((10 + roxl), (w1 + royl), abuf, strlen(abuf));
	}
	else
	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "AIA Covers Open");
		skp->Text((10 + roxl), (w1 + royl), bbuf, strlen(bbuf));
	}

	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "Target Longitude 102W (New Mexico)");
		skp->Text((10 + roxl), (w2 + royl), bbuf, strlen(bbuf));
	}

	if (mean_earth_long < 180)
	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "SDO Mean Surface Longitude %0.2fE", mean_earth_long);
		skp->Text((10 + roxl), (w3 + royl), bbuf, strlen(bbuf));
	}
	else
	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "SDO Mean Surface Longitude %0.2fW", 360 - mean_earth_long);
		skp->Text((10 + roxl), (w4 + royl), bbuf, strlen(bbuf));
	}

	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "Angle to Target Longitude %0.2fE (%0.2fW)", deg_long_to_target_e, deg_long_to_target_w);
		skp->Text((10 + roxl), (w5 + royl), bbuf, strlen(bbuf));
	}


	if (delta_long > 0)
	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "Drift Rate %0.3fE per orbit", delta_long);
		skp->Text((10 + roxl), (w6 + royl), bbuf, strlen(bbuf));
	}
	else
	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "Drift Rate %0.3fW per orbit", -delta_long);
		skp->Text((10 + roxl), (w6 + royl), bbuf, strlen(bbuf));
	}

	{
		skp->SetTextColor(0x0066FF66);
		char bbuf[256];
		sprintf(bbuf, "Orbits Required %0.3f", orbits_required);
		skp->Text((10 + roxl), (w7 + royl), bbuf, strlen(bbuf));
	}

	return true;
}

void sdo::clbkPostStep(double simt, double simdt, double mjd)
{
	stepsize = simdt;
	if (ant_status == 1) AntEarth();

	// Deploy hi-gain antenna
	if (ant_status >= DOOR_CLOSING) {
		double da = simdt * ANTENNA_DEPLOY_SPEED;
		if (ant_status == DOOR_CLOSING) {
			if (ant_proc > 0.0) ant_proc = max(0.0, ant_proc - da);
			else                ant_status = DOOR_CLOSED;
		}
		else {
			if (ant_proc < 1.0) ant_proc = min(1.0, ant_proc + da);
			else                ant_status = DOOR_OPEN;
		}
		SetAnimation(anim_ant, ant_proc);
	}


	// Deploy solpanels
	if (solp_status >= DOOR_CLOSING) {
		double da = simdt * SOLP_DEPLOY_SPEED;
		if (solp_status == DOOR_CLOSING) {
			if (solp_proc > 0.0) solp_proc = max(0.0, solp_proc - da);
			else                solp_status = DOOR_CLOSED;
		}
		else {
			if (solp_proc < 1.0) solp_proc = min(1.0, solp_proc + da);
			else                solp_status = DOOR_OPEN;
		}
		SetAnimation(anim_solp, solp_proc);
	}

	// Open lids
	if (lid_status >= DOOR_CLOSING) {
		double da = simdt * LID_DEPLOY_SPEED;
		if (lid_status == DOOR_CLOSING) {
			if (lid_proc > 0.0) lid_proc = max(0.0, lid_proc - da);
			else                lid_status = DOOR_CLOSED;
		}
		else {
			if (lid_proc < 1.0) lid_proc = min(1.0, lid_proc + da);
			else                lid_status = DOOR_OPEN;
		}
		SetAnimation(anim_lid, lid_proc);
	}


	DoCalculations();
}